@@ -4,7 +4,7 @@ Contains various `numpy` utility programs.
 
 import sys
 from math import cos, exp, floor, log, pi, sqrt
-from typing import Any, Callable, Iterable, cast
+from typing import Any, Callable, Iterable, Union, cast
 
 import numpy as np
 from numpy.polynomial import Polynomial
@@ -18,6 +18,142 @@ FourArrays = tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
 SixArrays = tuple[
     np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray
 ]
+FloatOrArray = Union[float, np.ndarray]
+Function_x = Callable[[np.ndarray], float]
+Function_xa = Callable[[np.ndarray, int], float]
+Function_xargs = Callable[[np.ndarray, list], float]
+ArrayFunctionOfArray = Callable[[np.ndarray], np.ndarray]
+
+
+def check_vector(v: Any, fun_name: str = None) -> int:
+    """
+    test that `v` is a vector; aborts otherwise
+
+    Args:
+        v: a vector, we hope
+        fun_name: name of the calling function
+
+    Returns:
+        the size if successful
+    """
+    fun_str = ["" if fun_name is None else fun_name + ":"]
+    if not isinstance(v, np.ndarray):
+        bs_error_abort(f"{fun_str} v should be a Numpy array")
+    v = cast(np.ndarray, v)
+    ndims_v = v.ndim
+    if ndims_v != 1:
+        bs_error_abort(f"{fun_str} v should have one dimension, not {ndims_v}")
+    return cast(int, v.size)
+
+
+def check_matrix(x: Any, fun_name: str = None) -> tuple[int, int]:
+    """
+    test that `x` is a matrix; aborts otherwise
+
+    Args:
+        x: a matrix, we hope
+        fun_name: name of the calling function
+
+    Returns:
+        the shape if successful
+    """
+    fun_str = ["" if fun_name is None else fun_name + ":"]
+    if not isinstance(x, np.ndarray):
+        bs_error_abort(f"{fun_str} Xx should be a Numpy array")
+    x = cast(np.ndarray, x)
+    ndims_x = x.ndim
+    if ndims_x != 2:
+        bs_error_abort(f"{fun_str} x should have two dimensions, not {ndims_x}")
+    return cast(tuple[int, int], x.shape)
+
+
+def check_vector_or_matrix(x: Any, fun_name: str = None) -> int:
+    """
+    test that `x` is a vector or a matrix; aborts otherwise
+
+    Args:
+        x: a vector or matrix, we hope
+        fun_name: name of the calling function
+
+    Returns:
+        the number of dimensions of `x` (1 or 2)
+    """
+    fun_str = ["" if fun_name is None else fun_name + ":"]
+    if not isinstance(x, np.ndarray):
+        bs_error_abort(f"{fun_str} X should be a Numpy array")
+    x = cast(np.ndarray, x)
+    ndims_x = x.ndim
+    if ndims_x != 1 and ndims_x != 2:
+        bs_error_abort(f"{fun_str} x should have at most two dimensions, not {ndims_x}")
+    return cast(int, ndims_x)
+
+
+def check_square(A: Any, fun_name: str = None) -> int:
+    """
+    test that an object used in `fun_name` is a square matrix
+
+    Args:
+        A: square matrix, we hope
+        fun_name: the name of the calling function
+
+    Returns:
+        the number of rows and columns of `A`
+    """
+    fun_str = ["" if fun_name is None else fun_name + ":"]
+    if not isinstance(A, np.ndarray):
+        bs_error_abort(f"{fun_str} A should be a Numpy array")
+    A = cast(np.ndarray, A)
+    if A.ndim == 2:
+        n, nv = A.shape
+        if nv != n:
+            bs_error_abort(f"{fun_str} The matrix A should be square, not {A.shape}")
+    else:
+        bs_error_abort(f"{fun_name} A should have  two dimensions, not {A.ndim}")
+    return cast(int, n)
+
+
+def check_tensor(x: Any, n_dims: int, fun_name: str = None) -> tuple[int, ...]:
+    """
+    test that `x` is an `n_dims` dimensional array; aborts otherwise
+
+    Args:
+        x: an `n_dims` dimensional array, we hope
+        fun_name: name of the calling function
+
+    Returns:
+        the shape if successful
+    """
+    fun_str = ["" if fun_name is None else fun_name + ":"]
+    if not isinstance(x, np.ndarray):
+        bs_error_abort(f"{fun_str} x should be a Numpy array")
+    x = cast(np.ndarray, x)
+    ndims_x = x.ndim
+    if ndims_x != n_dims:
+        bs_error_abort(f"{fun_str} x should have {n_dims} dimensions, not {ndims_x}")
+        return (0,)  # for mypy
+    return cast(tuple[int, ...], x.shape)
+
+
+def grid_function(
+    fun: Callable[[np.ndarray, np.ndarray], np.ndarray],
+    x_points: np.ndarray,
+    y_points: np.ndarray,
+) -> np.ndarray:
+    """apply a function f(x, y) on a lattice grid
+
+    Args:
+        fun: should return a matrix `(m, n)`  when called with two matrices `(m, n)`
+        x_points: an `m`-vector
+        y_points: an `n`-vector
+
+    Returns:
+        the `(m, n)` matrix of values of `fun` on the grid
+    """
+    _ = check_vector(x_points)
+    _ = check_vector(y_points)
+    X1, Y1 = np.meshgrid(x_points, y_points)
+    z_grid = fun(X1, Y1).T
+    return z_grid
 
 
 # Numpy parallel RNG
@@ -204,7 +340,7 @@ def nplog(
     verbose: bool = False,
 ) -> np.ndarray | TwoArrays | ThreeArrays:
     """
-    `C^2` extension of  `\\ln(a)` below `eps`, perhaps with derivatives
+    $C^2$ extension of  $\\ln(a)$ below `eps`, perhaps with derivatives
 
     Args:
         arr: any Numpy array
@@ -213,7 +349,7 @@ def nplog(
         verbose: prints debugging info
 
     Returns:
-        `\\ln(a)` `C^2`-extended below `eps`, perhaps with derivatives
+        $\\ln(a)$  $C^2$-extended below `eps`, perhaps with derivatives
     """
     if deriv not in [0, 1, 2]:
         bs_error_abort(f"deriv can only be 0, 1, or 2; not {deriv}")
@@ -260,7 +396,7 @@ def npexp(
     verbose: bool = False,
 ) -> np.ndarray | TwoArrays | ThreeArrays:
     """
-    `C^2` extension of  `\\exp(a)` above `bigx` and below `lowx`,
+    $C^2$ extension of  $\\exp(a)$ above `bigx` and below `lowx`,
     perhaps with derivatives
 
     Args:
@@ -272,7 +408,7 @@ def npexp(
 
 
     Returns:
-        `\\exp(a)`  `C^2`-extended above `bigx` and below `lowx`,
+        $\\exp(a)$  $C^2$-extended above `bigx` and below `lowx`,
         perhaps with derivatives
     """
     if deriv not in [0, 1, 2]:
@@ -483,81 +619,6 @@ def bsgrid(v: np.ndarray, w: np.ndarray) -> np.ndarray:
     v2 = np.tile(w, m)
     return np.column_stack((v1, v2))
 
-    """
-    This is a Python function that tests whether a given input is a vector and returns its size if
-    successful.
-    
-    :param v: `v` is a numpy array that is expected to be a vector
-    :type v: np.ndarray
-    :param fun_name: `fun_name` is an optional parameter that represents the name of the calling
-    function. If provided, it will be used in error messages to indicate which function caused the
-    error. If not provided, the error message will not include the function name
-    :type fun_name: Optional[str]
-    """
-
-
-def check_vector(v: Any, fun_name: str = None) -> int:
-    """
-    test that `v` is a vector; aborts otherwise
-
-    Args:
-        v: a vector, we hope
-        fun_name: name of the calling function
-
-    Returns:
-        the size if successful
-    """
-    fun_str = ["" if fun_name is None else fun_name + ":"]
-    if not isinstance(v, np.ndarray):
-        bs_error_abort(f"{fun_str} v should be a Numpy array")
-    v = cast(np.ndarray, v)
-    ndims_v = v.ndim
-    if ndims_v != 1:
-        bs_error_abort(f"{fun_str} v should have one dimension, not {ndims_v}")
-    return cast(int, v.size)
-
-
-def check_matrix(x: Any, fun_name: str = None) -> tuple[int, int]:
-    """
-    test that `x` is a matrix; aborts otherwise
-
-    Args:
-        x: a matrix, we hope
-        fun_name: name of the calling function
-
-    Returns:
-        the shape if successful
-    """
-    fun_str = ["" if fun_name is None else fun_name + ":"]
-    if not isinstance(x, np.ndarray):
-        bs_error_abort(f"{fun_str} Xx should be a Numpy array")
-    x = cast(np.ndarray, x)
-    ndims_x = x.ndim
-    if ndims_x != 2:
-        bs_error_abort(f"{fun_str} x should have two dimensions, not {ndims_x}")
-    return cast(tuple[int, int], x.shape)
-
-
-def check_vector_or_matrix(x: Any, fun_name: str = None) -> int:
-    """
-    test that `x` is a vector or a matrix; aborts otherwise
-
-    Args:
-        x: a vector or matrix, we hope
-        fun_name: name of the calling function
-
-    Returns:
-        the number of dimensions of `x` (1 or 2)
-    """
-    fun_str = ["" if fun_name is None else fun_name + ":"]
-    if not isinstance(x, np.ndarray):
-        bs_error_abort(f"{fun_str} X should be a Numpy array")
-    x = cast(np.ndarray, x)
-    ndims_x = x.ndim
-    if ndims_x != 1 and ndims_x != 2:
-        bs_error_abort(f"{fun_str} x should have at most two dimensions, not {ndims_x}")
-    return cast(int, ndims_x)
-
 
 def bs_sqrt_pdmatrix(m: np.ndarray) -> np.ndarray:
     """
@@ -578,52 +639,6 @@ def bs_sqrt_pdmatrix(m: np.ndarray) -> np.ndarray:
     return cast(np.ndarray, res)
 
 
-def check_square(A: Any, fun_name: str | None) -> int:
-    """
-    test that an object used in `fun_name` is a square matrix
-
-    Args:
-        A: square matrix, we hope
-        fun_name: the name of the calling function
-
-    Returns:
-        the number of rows and columns of `A`
-    """
-    fun_str = ["" if fun_name is None else fun_name + ":"]
-    if not isinstance(A, np.ndarray):
-        bs_error_abort(f"{fun_str} A should be a Numpy array")
-    A = cast(np.ndarray, A)
-    if A.ndim == 2:
-        n, nv = A.shape
-        if nv != n:
-            bs_error_abort(f"{fun_str} The matrix A should be square, not {A.shape}")
-    else:
-        bs_error_abort(f"{fun_name} A should have  two dimensions, not {A.ndim}")
-    return cast(int, n)
-
-
-def check_tensor(x: Any, n_dims: int, fun_name: str | None) -> tuple[int, ...]:
-    """
-    test that `x` is an `n_dims` dimensional array; aborts otherwise
-
-    Args:
-        x: an `n_dims` dimensional array, we hope
-        fun_name: name of the calling function
-
-    Returns:
-        the shape if successful
-    """
-    fun_str = ["" if fun_name is None else fun_name + ":"]
-    if not isinstance(x, np.ndarray):
-        bs_error_abort(f"{fun_str} x should be a Numpy array")
-    x = cast(np.ndarray, x)
-    ndims_x = x.ndim
-    if ndims_x != n_dims:
-        bs_error_abort(f"{fun_str} x should have {n_dims} dimensions, not {ndims_x}")
-        return (0,)  # for mypy
-    return cast(tuple[int, ...], x.shape)
-
-
 def make_lexico_grid(arr: np.ndarray) -> np.ndarray:
     """
     make a lexicographic grid
@@ -770,7 +785,7 @@ def npxlogx(
     verbose: bool = False,
 ) -> np.ndarray | TwoArrays | ThreeArrays:
     """
-    C^2` extension of  `a\\ln(a)` below `eps`, perhaps with derivatives
+    $C^2$ extension of  $a\\ln(a)$ below `eps`, perhaps with derivatives
 
     Args:
         arr: a Numpy array
@@ -779,7 +794,7 @@ def npxlogx(
         verbose: prints debugging info
 
     Returns:
-        `a\\ln(a)`  `C^2`-extended  below `eps`, perhaps with derivatives
+        $a\\ln(a)$  $C^2$-extended  below `eps`, perhaps with derivatives
     """
     if deriv not in [0, 1, 2]:
         bs_error_abort(f"deriv must be 0, 1, or 2; not {deriv}")
@@ -955,3 +970,91 @@ def gaussian_expectation(
                 integral_val += weights[i] * f(nodes[i], pars)
 
     return cast(np.ndarray, integral_vec) if vectorized else cast(float, integral_val)
+
+
+def legendre_polynomials(
+    x: np.ndarray,
+    max_deg: int,
+    a: float = -1.0,
+    b: float = 1.0,
+    no_constant: bool = False,
+) -> np.ndarray:
+    """evaluates the Legendre polynomials over x in the interval [a, b]
+
+    Args:
+        x: the points where the polynomials are to be evaluated
+        max_deg: the maximum degree
+        a: the start of the interval, classically -1
+        b: the end of the interval, classically 1
+        no_constant: if True, delete the constant polynomial
+
+    Returns:
+        an array of (max_deg+1) arrays of the shape of x
+    """
+    sx = check_vector(x)
+    if a > np.min(x):
+        sys.exit("legendre_polynomials: points below start of interval")
+    if b < np.max(x):
+        sys.exit("legendre_polynomials: points above end of interval")
+    p = np.zeros((sx, max_deg + 1))
+    x_transf = 2.0 * (x - a) / (b - a) - 1.0
+    p[:, 0] = np.ones_like(x)
+    p[:, 1] = x_transf
+    for deg in range(2, max_deg + 1):
+        p2 = (2 * deg - 1) * (p[:, deg - 1] * x_transf) - (deg - 1) * p[:, deg - 2]
+        p[:, deg] = p2 / deg
+    polys_p = p[:, 1:] if no_constant else p
+    return polys_p
+
+
+def quantile_transform(v: np.ndarray) -> np.ndarray:
+    """transform a vector of counts into the corresponding quantiles
+
+    Args:
+        v:  a vector of counts
+
+    Returns:
+         the corresponding quantiles
+    """
+    n = check_vector(v)
+    q = np.zeros(n)
+    for i in range(n):
+        q[i] = np.sum(v <= v[i]) / (n + 1)
+    return q
+
+
+def print_quantiles(
+    v: np.ndarray | Iterable[np.ndarray], quantiles: np.ndarray
+) -> np.ndarray:
+    """print these quantiles of the array(s)
+
+    Args:
+        v:  a vector or an iterable of vectors
+        quantiles: quantiles in [0,1]
+
+    Returns:
+         the corresponding quantiles as a vector or a matrix
+    """
+    nq = check_vector(quantiles)
+    if isinstance(v, np.ndarray):
+        qvals = np.quantile(v, quantiles)
+        for q, qv in zip(quantiles, qvals, strict=True):
+            print(f"Quantile {q: .3f}: {qv: >10.3f}")
+    elif isinstance(v, Iterable):
+        v = list(v)
+        for v_i in v:
+            _ = check_vector(v_i)
+        nv = len(v)
+        qvals = np.zeros((nq, nv))
+        for i in range(nv):
+            qvals[:, i] = np.quantile(v[i], quantiles)
+        for iq, q in enumerate(quantiles):
+            s = f"Quantile {q: .3f}: "
+            qv = qvals[iq, :]
+            for i in range(nv):
+                s += f"  {qv[i]: >10.3f}"
+            print(f"{s}")
+    else:
+        bs_error_abort("v must be  a vector or a list of vectors")
+
+    return cast(np.ndarray, qvals)
